@@ -77,9 +77,9 @@ read_compunits(Context<E> &ctx, ObjectFile<E> &file) {
   while (!data.empty()) {
     if (data.size() < 4)
       Fatal(ctx) << *file.debug_info << ": corrupted .debug_info";
-    if (*(ul32 *)data.data() == 0xffffffff)
+    if (*(U32<E> *)data.data() == 0xffff'ffff)
       Fatal(ctx) << *file.debug_info << ": --gdb-index: DWARF64 not supported";
-    i64 len = *(ul32 *)data.data() + 4;
+    i64 len = *(U32<E> *)data.data() + 4;
     vec.push_back(data.substr(0, len));
     data = data.substr(len);
   }
@@ -116,15 +116,15 @@ std::vector<GdbIndexName> read_pubnames(Context<E> &ctx, ObjectFile<E> &file) {
       if (contents.size() < 14)
         Fatal(ctx) << isec << ": corrupted header";
 
-      u32 len = *(ul32 *)contents.data() + 4;
-      u32 debug_info_offset = *(ul32 *)(contents.data() + 6);
+      u32 len = *(U32<E> *)contents.data() + 4;
+      u32 debug_info_offset = *(U32<E> *)(contents.data() + 6);
       u32 cu_idx = get_cu_idx(isec, debug_info_offset);
 
       std::string_view data = contents.substr(14, len - 14);
       contents = contents.substr(len);
 
       while (!data.empty()) {
-        u32 offset = *(ul32 *)data.data();
+        u32 offset = *(U32<E> *)data.data();
         data = data.substr(4);
         if (offset == 0)
           break;
@@ -176,7 +176,7 @@ static std::tuple<u8 *, u8 *, u32>
 find_compunit(Context<E> &ctx, ObjectFile<E> &file, i64 offset) {
   // Read .debug_info to find the record at a given offset.
   u8 *cu = get_buffer(ctx, ctx.debug_info) + offset;
-  u32 dwarf_version = *(ul16 *)(cu + 4);
+  u32 dwarf_version = *(U16<E> *)(cu + 4);
   u32 abbrev_offset;
 
   // Skip a header.
@@ -184,14 +184,14 @@ find_compunit(Context<E> &ctx, ObjectFile<E> &file, i64 offset) {
   case 2:
   case 3:
   case 4:
-    abbrev_offset = *(ul32 *)(cu + 6);
+    abbrev_offset = *(U32<E> *)(cu + 6);
     if (u32 address_size = cu[10]; address_size != sizeof(Word<E>))
       Fatal(ctx) << file << ": --gdb-index: unsupported address size "
                  << address_size;
     cu += 11;
     break;
   case 5: {
-    abbrev_offset = *(ul32 *)(cu + 8);
+    abbrev_offset = *(U32<E> *)(cu + 8);
     if (u32 address_size = cu[7]; address_size != sizeof(Word<E>))
       Fatal(ctx) << file << ": --gdb-index: unsupported address size "
                  << address_size;
@@ -295,7 +295,7 @@ public:
 // Read value of the given DW_FORM_* form. If a value is not scalar,
 // returns a dummy value 0.
 template <typename E>
-inline u64 DebugInfoReader<E>::read(u64 form) {
+u64 DebugInfoReader<E>::read(u64 form) {
   switch (form) {
   case DW_FORM_flag_present:
     return 0;
@@ -309,13 +309,13 @@ inline u64 DebugInfoReader<E>::read(u64 form) {
   case DW_FORM_strx2:
   case DW_FORM_addrx2:
   case DW_FORM_ref2: {
-    u64 val = *(ul16 *)cu;
+    u64 val = *(U16<E> *)cu;
     cu += 2;
     return val;
   }
   case DW_FORM_strx3:
   case DW_FORM_addrx3: {
-    u64 val = *(ul24 *)cu;
+    u64 val = *(U24<E> *)cu;
     cu += 3;
     return val;
   }
@@ -326,13 +326,13 @@ inline u64 DebugInfoReader<E>::read(u64 form) {
   case DW_FORM_strx4:
   case DW_FORM_addrx4:
   case DW_FORM_ref4: {
-    u64 val = *(ul32 *)cu;
+    u64 val = *(U32<E> *)cu;
     cu += 4;
     return val;
   }
   case DW_FORM_data8:
   case DW_FORM_ref8: {
-    u64 val = *(ul64 *)cu;
+    u64 val = *(U64<E> *)cu;
     cu += 8;
     return val;
   }
@@ -359,8 +359,7 @@ inline u64 DebugInfoReader<E>::read(u64 form) {
   }
 }
 
-// Read a range list from .debug_ranges starting at the given offset
-// (until an end of list entry).
+// Read a range list from .debug_ranges starting at the given offset.
 template <typename E>
 static std::vector<u64>
 read_debug_range(Context<E> &ctx, ObjectFile<E> &file, Word<E> *range) {
@@ -379,8 +378,7 @@ read_debug_range(Context<E> &ctx, ObjectFile<E> &file, Word<E> *range) {
   return vec;
 }
 
-// Read a range list from .debug_rnglists starting at the given offset
-// (until an end of list entry).
+// Read a range list from .debug_rnglists starting at the given offset.
 template <typename E>
 static std::vector<u64>
 read_rnglist_range(Context<E> &ctx, ObjectFile<E> &file, u8 *rnglist,
@@ -501,7 +499,7 @@ read_address_areas(Context<E> &ctx, ObjectFile<E> &file, i64 offset) {
       Fatal(ctx) << file << ": --gdb-index: missing DW_AT_rnglists_base";
 
     u8 *base = buf + *rnglists_base;
-    return read_rnglist_range(ctx, file, base + *(ul32 *)base, addrx);
+    return read_rnglist_range(ctx, file, base + *(U32<E> *)base, addrx);
   }
 
   // Handle a contiguous address range.
@@ -546,16 +544,11 @@ read_address_areas(Context<E> &ctx, ObjectFile<E> &file, i64 offset) {
   return {};
 }
 
-#define INSTANTIATE(E)                                                  \
-  template std::vector<std::string_view>                                \
-  read_compunits(Context<E> &, ObjectFile<E> &);                        \
-  template std::vector<GdbIndexName>                                    \
-  read_pubnames(Context<E> &, ObjectFile<E> &);                         \
-  template i64                                                          \
-  estimate_address_areas(Context<E> &, ObjectFile<E> &);                \
-  template std::vector<u64>                                             \
-  read_address_areas(Context<E> &, ObjectFile<E> &, i64)
+using E = MOLD_TARGET;
 
-INSTANTIATE_ALL;
+template std::vector<std::string_view> read_compunits(Context<E> &, ObjectFile<E> &);
+template std::vector<GdbIndexName> read_pubnames(Context<E> &, ObjectFile<E> &);
+template i64 estimate_address_areas(Context<E> &, ObjectFile<E> &);
+template std::vector<u64> read_address_areas(Context<E> &, ObjectFile<E> &, i64);
 
 } // namespace mold::elf
