@@ -1,8 +1,7 @@
 #pragma once
 
-#include "../integers.h"
+#include "../common/integers.h"
 
-#include <cstdint>
 #include <ostream>
 #include <string>
 #include <type_traits>
@@ -1899,7 +1898,6 @@ struct PPC64V1 {
   static constexpr u32 page_size = 65536;
   static constexpr u32 e_machine = EM_PPC64;
   static constexpr u32 plt_hdr_size = 52;
-  static constexpr u32 plt_size = 8;
   static constexpr u32 pltgot_size = 0;
   static constexpr u32 thunk_hdr_size = 0;
   static constexpr u32 thunk_size = 28;
