@@ -239,7 +239,8 @@ void parse_linker_script(Context<E> &ctx, MappedFile<Context<E>> *mf) {
 }
 
 template <typename E>
-MachineType get_script_output_type(Context<E> &ctx, MappedFile<Context<E>> *mf) {
+std::string_view
+get_script_output_type(Context<E> &ctx, MappedFile<Context<E>> *mf) {
   current_file<E> = mf;
 
   std::vector<std::string_view> vec = tokenize(ctx, mf->get_contents());
@@ -247,9 +248,9 @@ MachineType get_script_output_type(Context<E> &ctx, MappedFile<Context<E>> *mf)
 
   if (tok.size() >= 3 && tok[0] == "OUTPUT_FORMAT" && tok[1] == "(") {
     if (tok[2] == "elf64-x86-64")
-      return MachineType::X86_64;
+      return X86_64::target_name;
     if (tok[2] == "elf32-i386")
-      return MachineType::I386;
+      return I386::target_name;
   }
 
   if (tok.size() >= 3 && (tok[0] == "INPUT" || tok[0] == "GROUP") &&
@@ -258,7 +259,7 @@ MachineType get_script_output_type(Context<E> &ctx, MappedFile<Context<E>> *mf)
         MappedFile<Context<E>>::open(ctx, std::string(unquote(tok[2]))))
       return get_machine_type(ctx, mf);
 
-  return MachineType::NONE;
+  return "";
 }
 
 static bool read_label(std::span<std::string_view> &tok,
@@ -311,7 +312,7 @@ read_version_script_commands(Context<E> &ctx, std::span<std::string_view> &tok,
     }
 
     if (tok[0] == "*") {
-      ctx.default_version = (is_global ? ver_idx : VER_NDX_LOCAL);
+      ctx.default_version = (is_global ? ver_idx : (u32)VER_NDX_LOCAL);
       ctx.default_version_from_version_script = true;
     } else if (is_global) {
       ctx.version_patterns.push_back({unquote(tok[0]), current_file<E>->name,
@@ -416,7 +417,7 @@ void parse_dynamic_list(Context<E> &ctx, MappedFile<Context<E>> *mf) {
 using E = MOLD_TARGET;
 
 template void parse_linker_script(Context<E> &, MappedFile<Context<E>> *);
-template MachineType get_script_output_type(Context<E> &, MappedFile<Context<E>> *);
+template std::string_view get_script_output_type(Context<E> &, MappedFile<Context<E>> *);
 template void parse_version_script(Context<E> &, MappedFile<Context<E>> *);
 template void parse_dynamic_list(Context<E> &, MappedFile<Context<E>> *);
 
