@@ -1,6 +1,6 @@
-#!/bin/bash
+#!/bin/sh
 set -e
-source /etc/os-release
+. /etc/os-release
 
 set -x
 
@@ -40,6 +40,11 @@ arch-*)
   pacman -Sy
   pacman -S --needed --noconfirm base-devel zlib openssl cmake util-linux
   ;;
+void-*)
+  xbps-install -Sy xbps
+  xbps-install -Sy bash make cmake openssl-devel zlib-devel gcc
+  xbps-install -Sy tar diffutils util-linux
+  ;;
 *)
   echo "Error: don't know anything about build dependencies on $ID-$VERSION_ID"
   exit 1
