@@ -1,22 +1,14 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
-[ $MACHINE = x86_64 ] || { echo skipped; exit; }
+# This test crashes only on qemu-sparc64 running on GitHub Actions,
+# even though it works on a local x86-64 machine and on an actual
+# SPARC machine.
+[ $MACHINE = sparc64 ] && skip
 
 cat <<EOF | $CC -o $t/a.o -c -x assembler -
 .globl foo
-foo = 0x800000
+foo = 0x800008
 EOF
 
 cat <<EOF | $CC -o $t/b.o -c -fno-PIC -xc -
@@ -27,12 +19,11 @@ cat <<EOF | $CC -o $t/b.o -c -fno-PIC -xc -
 #include <ucontext.h>
 
 void handler(int signum, siginfo_t *info, void *ptr) {
-  ucontext_t *u = (ucontext_t *)ptr;
-  printf("ip=0x%llx\n", u->uc_mcontext.gregs[REG_RIP]);
+  printf("ip=%p\n", info->si_addr);
   exit(0);
 }
 
-void foo();
+extern volatile int foo;
 
 int main() {
   struct sigaction act;
@@ -40,11 +31,9 @@ int main() {
   act.sa_sigaction = handler;
   sigemptyset(&act.sa_mask);
   sigaction(SIGSEGV, &act, 0);
-  foo();
+  foo = 5;
 }
 EOF
 
 $CC -B. -o $t/exe -no-pie $t/a.o $t/b.o
-$QEMU $t/exe | grep -q '^ip=0x800000$'
-
-echo OK
+$QEMU $t/exe | grep -q '^ip=0x80000.$'
