@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 echo 'int main() { return 0; }' > $t/a.c
 
@@ -32,5 +21,3 @@ readelf -n $t/exe | grep -q 'GNU.*0x00000020.*NT_GNU_BUILD_ID'
 
 $CC -B. -o $t/exe $t/a.c -Wl,-build-id=0xdeadbeefdeadbeef
 readelf -n $t/exe | grep -q 'Build ID: deadbeefdeadbeef'
-
-echo OK
