@@ -1,18 +1,11 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
-[ $MACHINE = riscv64 -o $MACHINE = riscv32 ] && { echo skipped; exit; }
+[ $MACHINE = riscv64 -o $MACHINE = riscv32 ] && skip
+
+# musl libc does not support init/fini on ARM
+# https://github.com/rui314/mold/issues/951
+[ $MACHINE = arm -o $MACHINE = aarch64 ] && ldd --help 2>&1 | grep -q musl && skip
 
 cat <<EOF | $CC -c -fPIC -o $t/a.o -xc -
 void keep();
@@ -39,8 +32,8 @@ EOF
 $CC -B. -o $t/c.so -shared $t/b.o
 $CC -B. -o $t/d.so -shared $t/b.o -Wl,-init,init -Wl,-fini,fini
 
-$CC -o $t/exe1 $t/a.o $t/c.so
-$CC -o $t/exe2 $t/a.o $t/d.so
+$CC -B. -o $t/exe1 $t/a.o $t/c.so
+$CC -B. -o $t/exe2 $t/a.o $t/d.so
 
 $QEMU $t/exe1 > $t/log1
 $QEMU $t/exe2 > $t/log2
@@ -50,5 +43,3 @@ $QEMU $t/exe2 > $t/log2
 
 grep -q init $t/log2
 grep -q fini $t/log2
-
-echo OK
