@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 cat <<EOF | $CC -fPIC -xc -c -o $t/a.o -
 int foo() {
@@ -43,26 +32,24 @@ EOF
 
 $CC -B. -shared -o $t/f.so $t/e.o $t/c.a $t/d.a
 readelf --dyn-syms $t/f.so > $t/log
-fgrep -q foo $t/log
-fgrep -q bar $t/log
-fgrep -q baz $t/log
+grep -Fq foo $t/log
+grep -Fq bar $t/log
+grep -Fq baz $t/log
 
 $CC -B. -shared -o $t/f.so $t/e.o $t/c.a $t/d.a -Wl,-exclude-libs=c.a
 readelf --dyn-syms $t/f.so > $t/log
-! fgrep -q foo $t/log || false
-fgrep -q bar $t/log
-fgrep -q baz $t/log
+! grep -Fq foo $t/log || false
+grep -Fq bar $t/log
+grep -Fq baz $t/log
 
 $CC -B. -shared -o $t/f.so $t/e.o $t/c.a $t/d.a -Wl,-exclude-libs=c.a -Wl,-exclude-libs=d.a
 readelf --dyn-syms $t/f.so > $t/log
-! fgrep -q foo $t/log || false
-! fgrep -q bar $t/log || false
-fgrep -q baz $t/log
+! grep -Fq foo $t/log || false
+! grep -Fq bar $t/log || false
+grep -Fq baz $t/log
 
 $CC -B. -shared -o $t/f.so $t/e.o $t/c.a $t/d.a -Wl,-exclude-libs=ALL
 readelf --dyn-syms $t/f.so > $t/log
-! fgrep -q foo $t/log || false
-! fgrep -q bar $t/log || false
-fgrep -q baz $t/log
-
-echo OK
+! grep -Fq foo $t/log || false
+! grep -Fq bar $t/log || false
+grep -Fq baz $t/log
