@@ -1,16 +1,7 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
+
+[ $MACHINE = sh4 ] && skip
 
 cat <<EOF | $CXX -o $t/a.o -c -xc++ -
 #include <iostream>
@@ -36,5 +27,3 @@ EOF
 
 $CXX -B. -o $t/exe $t/a.o
 $QEMU $t/exe | grep -q 'error: exception'
-
-echo OK
