@@ -1,8 +1,8 @@
 #!/bin/bash
 . $(dirname $0)/common.inc
 
-seq 1 100000 | sed 's/.*/int x\0 __attribute__((section(".data.\0")));/g' | \
-  $CC -c -xc -o $t/a.o -
+seq 1 100000 | sed 's/.*/.section .data.\0,"aw"\n.word 0\n/g' | \
+  $CC -c -xassembler -o $t/a.o -
 
 cat <<'EOF' | $CC -c -xc -o $t/b.o -
 #include <stdio.h>
