@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 cat <<EOF | $CC -o $t/a.so -fPIC -shared -xc -
 void *foo() {
@@ -47,5 +36,3 @@ $QEMU $t/exe | grep -q '^1 1 1$'
 
 readelf --dyn-syms $t/exe | grep -q '00000000 .* foo'
 readelf --dyn-syms $t/exe | grep -q '00000000 .* bar'
-
-echo OK
