@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 cat <<EOF | $CC -o $t/a.o -ffunction-sections -c -xc -
 #include <stdio.h>
@@ -31,5 +20,3 @@ $CC -B. -o $t/exe2 $t/a.o $t/b.o -Wl,-shuffle-sections
 $QEMU $t/exe2 | grep -q 'Hello world'
 
 ! diff $t/exe1 $t/exe2 >& /dev/null || false
-
-echo OK
