@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 cat <<'EOF' | $CC -c -o $t/a.o -xc -
 __attribute__((section("foo")))
@@ -36,5 +25,3 @@ $QEMU $t/exe | grep -q 'section foo section foo'
 
 $CC -B. -o $t/exe $t/c.o $t/b.a -Wl,-gc-sections
 $QEMU $t/exe | grep -q 'section foo section foo'
-
-echo OK
