@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 cat <<EOF | $CC -o $t/a.o -c -x assembler -
   .globl foo, bar, this_is_global
@@ -41,5 +30,3 @@ grep -Eq '0 NOTYPE  LOCAL  DEFAULT .* module_local' $t/log
 grep -Eq '0 NOTYPE  GLOBAL DEFAULT .* foo' $t/log
 grep -Eq '0 NOTYPE  GLOBAL DEFAULT .* bar' $t/log
 grep -Eq '0 NOTYPE  GLOBAL DEFAULT .* this_is_global' $t/log
-
-echo OK
