@@ -1,24 +1,12 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
-if [ $MACHINE = x86_64 ]; then
+if [ $MACHINE = x86_64 -o $MACHINE = arm ]; then
   dialect=gnu2
 elif [ $MACHINE = aarch64 ]; then
   dialect=desc
 else
-  echo skipped
-  exit
+  skip
 fi
 
 cat <<EOF | $GCC -fPIC -mtls-dialect=$dialect -c -o $t/a.o -xc -
@@ -40,5 +28,3 @@ EOF
 
 $CC -B. -o $t/exe $t/a.o $t/b.so
 $QEMU $t/exe | grep -q '5 7'
-
-echo OK
