@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 cat <<EOF | $CC -c -o $t/a.o -xc -
 #include <stdio.h>
@@ -45,5 +34,3 @@ $CC -B. -o $t/exe $t/a.o $t/b.o $t/c.so \
 grep -q 'trace-symbol: .*/a.o: reference to foo' $t/log
 grep -q 'trace-symbol: .*/b.o: definition of foo' $t/log
 grep -q 'trace-symbol: .*/c.so: definition of baz' $t/log
-
-echo OK
