@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 cat <<'EOF' > $t/a.ver
 {
@@ -40,5 +29,3 @@ grep -q ' foobarcx$' $t/log
 grep -q ' foo123bar456bx$' $t/log
 ! grep -q ' foo123bar456c$' $t/log || false
 ! grep -q ' foo123bar456x$' $t/log || false
-
-echo OK
