@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 cat <<EOF | $CC -xc -c -o $t/a.o -
 __attribute__((visibility("hidden"))) int foo = 3;
@@ -31,6 +20,4 @@ EOF
 
 $CC -B. -shared -o $t/e.so $t/c.a $t/d.o
 readelf --dyn-syms $t/e.so > $t/log
-! fgrep -q foo $t/log || false
-
-echo OK
+! grep -Fq foo $t/log || false
